// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// CR x7 jcv 19-Jan-09 enhanced tracing
// SPR x24 12-Feb-2010 enhanced tracing again

#include "Pass.h"
#include "errstr.h"
#include "math.h"
#include "nrutil.h"

#include <iostream>

bool mrqmin(Vec_I_DP &x, Vec_I_DP &y, Vec_I_DP &sig, Vec_IO_DP &a, 
            Vec_I_BOOL &ia, Mat_O_DP &covar, Mat_O_DP &alpha, DP &chisq, 
	    void funcs(const DP, Vec_I_DP &, DP &, Vec_O_DP &), 
	    DP &alamda);

void mrqcof(Vec_I_DP &x, Vec_I_DP &y, Vec_I_DP &sig, Vec_I_DP &a, 
            Vec_I_BOOL &ia, Mat_O_DP &alpha, Vec_O_DP & beta, DP &chisq, 
	    void funcs(const DP, Vec_I_DP &, DP &, Vec_O_DP &));

void covsrt(Mat_IO_DP &covar, Vec_I_BOOL &ia, const int mfit);

bool gaussj(Mat_IO_DP &a, Mat_IO_DP &b);

void fit_eval(const DP x, Vec_I_DP &a, DP &y, Vec_O_DP &dyda);
  
Pass::Pass(int gs, int passno, int s) : GSID(gs), passNumber(passno), Start(s)  {

  // CR x7
  std::cout << "Pass::Pass , creating new instance for Pass number " << passno << std::endl;

  inRev = false;
  fitCalculated = false;
  rms = 0.0;
  gradient = 0.0;
  constant = 0.0;

  dataStartOffset = 0;
  
}
  
void Pass::addPoint(double obt, double ert){
  
  ERT.push_back(ert);
  OBT.push_back(obt);
  sigma.push_back(1.0e-6);
  
}

double Pass::getPredictedERT(double obt){

  double ert = 0.0;

  if ( ERT.size() == 0 ) {    
    ert = 0.0;
  }      

  if (!fitCalculated) {
    // SPRx24
    std::cout << "Pass:: called getPredicted but not fit yet. Fitting NOW" << std::endl;
    std::cout << "Pass:: OBT size is " << OBT.size() << std::endl;
    calculateFit();
    // SPRx24
    std::cout << "Pass:: Fit (constant gradient and rms) for pass already available." << std::endl;
  }

  ert = constant + (gradient * obt);
  return ert;

}

void Pass::calculateFit() {

  // CR x7
  std::cout << "Pass::calculateFit :  calling MRQFit" << std::endl;

  // If mrqmin fit fails, use a boring old LS Fit
  if (!calculateMRQFit(dataStartOffset)) {
    errstr << "MRQ Fit failed, using LS Fit" << Sas::warning("FitFailed");
    calculateLSFit(dataStartOffset);
  }
  
  double sumSquareMean = 0.0;
  double error;

  for (unsigned int i = 0; i < OBT.size(); i++) {

    error = ERT[i] - (constant + (gradient * OBT[i]));
    sumSquareMean += error * error;

  }

  rms = sqrt(sumSquareMean/OBT.size());
  // CR x7
  std::cout << "Pass::calculateFit -- constant : gradient : rms " << std::endl;

  std::cout << constant << " : " << gradient << " : " << rms << std::endl;

}

void Pass::calculateLSFit(int startOffset) {
  
  int Start;

  // CR x7
  std::cout << "Pass::calculateLSFit : " << std::endl;
  std::cout << "ERT.size : startOffset" << std::endl;
  std::cout <<  ERT.size() << ":" << startOffset << std::endl;

  if (startOffset != 0) {
    Start = ERT.size() - startOffset;
  }
  else {
    Start = 0;
  }

  int End   = ERT.size();

  // CR x7
  std::cout << "Start : End " << std::endl;
  std::cout << Start << ":" << End << std::endl;

  long double sumX = 0.0;
  long double sumY = 0.0;
  long double sumoffX2 = 0.0;
  long double avgX = 0.0;
  long double offX = 0.0;

  int nPoints = 0;
  int i;

  for (i = Start; i < End; i++) {
      
    sumX += OBT[i];
    sumY += ERT[i];
    
    nPoints++;

  }

  // Calculate average X value and initilise variable for
  // calculation of gradient and constant.
  
  avgX     = sumX/(long double)nPoints;
  sumoffX2 = 0.0;
  
  for (i = Start; i < End; i++) {
    offX      = OBT[i] - avgX;
    sumoffX2 += offX * offX;
    gradient += offX * ERT[i];
  }

  // CR x7
  std::cout << "sumY : sumX : sumoffX2 : nPoints " << std::endl;
  std::cout << sumY << ":" << sumX << ":" << sumoffX2 << ":" << nPoints << std::endl;
  
  // Calculate the gradient and constant.
  gradient /= sumoffX2;
  constant = (sumY - (sumX * gradient))/(long double)nPoints;

  fitCalculated = true;

}

bool Pass::calculateMRQFit(int startOffset) {

  int Start = 0;
  int End   = ERT.size();

  if (startOffset != 0) {
    Start = End - startOffset;
  }

  int dataSize = End - Start;

  Vec_DP x(dataSize);
  Vec_DP y(dataSize);
  Vec_DP sig(dataSize);

  for (int i = Start, j = 0; i < End; i++, j++) {
    x[j] = OBT[i];
    y[j] = ERT[i];
    sig[j] = sigma[i];
  }

  Vec_DP a(2);
  Vec_BOOL lista(2);
  Mat_DP covar(2,2);
  Mat_DP alpha(2,2);

  DP chisq, alamda;

  lista[0] = true;
  lista[1] = true;
  a[0] = 0.0;
  a[1] = 1.0;
  alamda = -1.0;
  chisq = 0.0;

  double chiold = 0.0;
  for (int i = 0; i < 50; i++) {

    // Run the fit
    if (!mrqmin(x,y,sig,a,lista,covar,alpha,chisq,fit_eval,alamda)) {
      fitCalculated = false;
      return false;
    }
    
    // Have we converged
    if (i == 0) {
      chiold = chisq;
    }
    else {
      if ((chiold-chisq)/chiold < 5e-8) {
	break;
      }
      chiold = chisq;
    }
  } 

  // Get the matrices
  alamda = 0.0;
  mrqmin(x,y,sig,a,lista,covar,alpha,chisq,fit_eval,alamda);
 
  constant = a[0];
  gradient = a[1];

  if (constant == 0.0 && gradient == 1.0) {

    fitCalculated = false;
    return false;

  }

  fitCalculated = true;
  return true;

}

void fit_eval(const DP x, Vec_I_DP &a, DP &y, Vec_O_DP &dyda) {

  DP y1;
  int na=a.size();

  y = 0;

  Vec_DP p(2);
  Vec_DP pa(2);

  p[0] = a[0];
  p[1] = a[1];

  y1 = p[0] + p[1] * x;
  y = y + y1;

  if (na > 0) {
    dyda[0] = 1.0;
    dyda[1] = x;
  }

}

/*---------------------------------------------------------------------------*/
/* MRQMIN routine, adapted from Numerical Recipes, 2nd ed. */
bool mrqmin(Vec_I_DP &x, Vec_I_DP &y, Vec_I_DP &sig, Vec_IO_DP &a, 
            Vec_I_BOOL &ia, Mat_O_DP &covar, Mat_O_DP &alpha, DP &chisq, 
	    void funcs(const DP, Vec_I_DP &, DP &, Vec_O_DP &), 
	    DP &alamda){

  int j, k, l;
  static int mfit;
  static DP ochisq;

  int ma=a.size();
  static Mat_DP *oneda_p;
  static Vec_DP *atry_p,*beta_p,*da_p;
  if (alamda < 0.0) {
    atry_p = new Vec_DP(ma);
    beta_p = new Vec_DP(ma);
    da_p = new Vec_DP(ma);
    mfit=0;
    
    for (j=0; j<ma; j++) if (ia[j]) mfit++;
    oneda_p = new Mat_DP(mfit,1);
    alamda=0.001;
    mrqcof(x,y,sig,a,ia,alpha,*beta_p,chisq,funcs);
    ochisq=chisq;
    for (j=0;j<ma;j++) (*atry_p)[j]=a[j];
  }
  Mat_DP &oneda=*oneda_p;
  Vec_DP &atry=*atry_p,&beta=*beta_p,&da=*da_p;
  Mat_DP temp(mfit,mfit);

  for (j=0; j<mfit; j++) {
    for (k=0; k<mfit; k++) covar[j][k] = alpha[j][k];
    covar[j][j] = alpha[j][j]*(1.0 + alamda);
    for (k=0; k<mfit; k++) temp[j][k] = covar[j][k];
    oneda[j][0] = beta[j];
  }

  if (!gaussj(temp, oneda)) {
    delete oneda_p; delete da_p; delete beta_p; delete atry_p;
    return false;
  }

  for (j=0; j<mfit; j++) {
    for (k=0; k<mfit; k++) covar[j][k] = temp[j][k];
    da[j] = oneda[j][0];
  }

  if (alamda == 0.) {
    covsrt(covar, ia, mfit);
    covsrt(alpha, ia, mfit);
    delete oneda_p; delete da_p; delete beta_p; delete atry_p;
    return true;
  }

  for (j=0, l=0; l<ma; l++) 
    if (ia[l]) atry[l] = a[l] + da[j++];

  mrqcof(x, y, sig, atry, ia, covar, da, chisq, funcs);

  if (chisq < ochisq) {
    alamda *= 0.1;
    ochisq = chisq;

    for (j=0; j<mfit; j++) {
      for (k=0; k<mfit; k++) alpha[j][k] = covar[j][k];
      beta[j] = da[j];
    }
    for (l=0; l<ma; l++) a[l] = atry[l];

  } else {
    alamda *= 10.0;
    chisq = ochisq;
  }

  return true;

}


/*---------------------------------------------------------------------------*/
/* MRQCOF routine, adapted from Numerical Recipes, 2nd ed. */
void mrqcof(Vec_I_DP &x, Vec_I_DP &y, Vec_I_DP &sig, Vec_I_DP &a, 
            Vec_I_BOOL &ia, Mat_O_DP &alpha, Vec_O_DP & beta, DP &chisq, 
	    void funcs(const DP, Vec_I_DP &, DP &, Vec_O_DP &)){

  int i, j, k, l, m, mfit=0;
  DP ymod, wt, sig2i, dy;

  int ndata = x.size();
  int ma = a.size();
  
  Vec_DP dyda(ma);

  for (j=0; j<ma; j++) if (ia[j]) mfit++;
  for (j=0; j<mfit; j++) {
    for (k=0; k<=j; k++) alpha[j][k] = 0.0;
    beta[j] = 0.0;
  }

  chisq = 0.0;

  for (i=0; i<ndata; i++) {
    funcs(x[i], a, ymod, dyda);
    sig2i = 1.0/(sig[i]*sig[i]);
    dy = y[i] - ymod;
    for (j=0, l=0; l<ma; l++) {
      if (ia[l]) {
	wt = dyda[l]*sig2i;
	for (k=0, m=0; m<l+1; m++)
	  if (ia[m]) alpha[j][k++] += wt*dyda[m];
	beta[j++] += dy*wt;
      }
    }
    chisq += dy*dy*sig2i;
  }

  for (j=1; j<mfit; j++)
    for (k=0; k<j; k++) alpha[k][j] = alpha[j][k];

}

/*---------------------------------------------------------------------------*/
/* COVSRT routine, adapted from Numerical Recipes, 2nd ed. */
void covsrt(Mat_IO_DP &covar, Vec_I_BOOL &ia, const int mfit){
  int i, j, k;

  int ma=ia.size(); 
  for (i=mfit; i<ma; i++)
    for (j=0; j<i+1; j++) covar[i][j] = covar[j][i] = 0.0;
  k = mfit - 1;
  for (j=ma - 1; j>=0; j--) {
    if (ia[j]) {
      for (i=0; i<ma; i++) SWAP(covar[i][k], covar[i][j]);
      for (i=0; i<ma; i++) SWAP(covar[k][i], covar[j][i]);
      k--;
    }
  }
}

/*---------------------------------------------------------------------------*/
/* GAUSSJ routine, adapted from Numerical Recipes, 2nd ed. */
bool gaussj(Mat_IO_DP &a, Mat_IO_DP &b){

  int i, icol=0, irow=0, j, k, l, ll;
  DP big, dum, pivinv;

  int n=a.nrows();
  int m=b.ncols();

  Vec_INT indxc(n), indxr(n), ipiv(n);
  
  for (j=0; j<n; j++) ipiv[j]=0;
  for (i=0; i<n; i++) {
    big = 0.0;
    for (j=0; j<n; j++) {
      if (ipiv[j] != 1) {
	for (k=0; k<n; k++) {
	  if (ipiv[k] == 0) {
	    if (fabs(a[j][k]) >= big) {
	      big = fabs(a[j][k]);
	      irow = j;
	      icol = k;
	    }
	  } 
	  else {
	    //	    if (ipiv[k] > 1) return false;
	  }
	}
      }
    }
    ++(ipiv[icol]);

    if (irow != icol) {
      for (l=0; l<n; l++) SWAP(a[irow][l], a[icol][l]);
      for (l=0; l<m; l++) SWAP(b[irow][l], b[icol][l]);
    }

    indxr[i] = irow;
    indxc[i] = icol;
    if (a[icol][icol] == 0.0) return false;

    pivinv = 1.0/a[icol][icol];
    a[icol][icol] = 1.0;

    for (l=0; l<n; l++) a[icol][l] *= pivinv;
    for (l=0; l<m; l++) b[icol][l] *= pivinv;

    for (ll=0; ll<n; ll++)
      if (ll != icol) {
	dum = a[ll][icol];
	a[ll][icol] = 0.0;
	for (l=0; l<n; l++) a[ll][l] -= a[icol][l]*dum;
	for (l=0; l<m; l++) b[ll][l] -= b[icol][l]*dum;
      }
  }
  for (l=n-1; l>=0; l--) {
    if (indxr[l] != indxc[l])
      for (k=0; k<n; k++)
	SWAP(a[k][indxr[l]], a[k][indxc[l]]);
  }

  return true;

}
