// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// History:
// CR x2 jcv 8-Oct-08 New keywords to be written in RTX file
// SPR x2 jcv 10-Oct-08 Delay keyword should be read from file itself 
// SPR x3 jcv 13-Oct-08 Versions should run as 001 instead of 0001
// CR x8 jcv 19-Jan-09 New plots for debugging
// SRPx15 jcv 22-May-09 Fixed passes verbosity
// SPRx24 jcv 12-FEb-2010 enhanced verbosity
// SPRx27 jcv 15-Feb-2010 No scatter aiming better plots 
// SPRx31 jcv 24-Mar-2010 Added moveup function for new columns

#include "RTX.h"
#include "Pass.h"

#include "Utils.h"

#include "dsplot.h"
#include "param.h"
#include "CalOalUtility.h"

#include <sstream>
#include <vector>
#include <map>

RTX::RTX(STime Start, STime End, int rev, int ver, int size) {
 
  // CR x2 New keywords
  // SPR x2 Fixes 
  string sTcxVer = "TCXVER";
  //string sDelayVer = "DELAYVER";
  //string sDelayVer;
  string sDelayVer;
  string sDelaysFilename;

  stringstream sDelayStream;
  // SPR x3
  //sDelayStream.width(4); sDelayStream.fill('0');
  sDelayStream.width(3); sDelayStream.fill('0');
 
  revolution = rev;
  version = ver;
  start = Start;
  end = End;

  // Set the revolution as a string
  stringstream revString; 
  revString.width(4); revString.fill('0');
  revString << revolution;

  stringstream verString; 
  // SPR x3
  //verString.width(4); verString.fill('0');
  verString.width(3); verString.fill('0');
  verString << version;

  // Set the size of the ASCII fields
  vector<unsigned long>   timesize;
  timesize.push_back( 27 );

  vector<unsigned long>   gsidsize;
  gsidsize.push_back( 5 );
  
  dataSet = dataSetServer->open( getFilename(".FIT"), DataSet::Create );
  
  // Set the top-level attributes
  // AI 2008_02_21: Added FILENAME keyword to be changed later on by qcheck
  *dataSet->addAttribute("FILENAME", "Filename") =getFilename(".FIT") ;
  *dataSet->addAttribute("ORBIT", "XMM Revolution") = revString.str();
  *dataSet->addAttribute("VERSION", "Version of this datafile") = verString.str();

  // CR x2 New keywords to be written in RTX file
  *dataSet->addAttribute("TCXVER","Version of TCX production software") = std::getenv(sTcxVer.c_str()); 

  // SPR x2 Fixes delay keywords 
  //*dataSet->addAttribute("DELAYVER","Version of DELAY.fit file") = std::getenv(sDelayVer.c_str()); 

  if(std::getenv("TDLYFILENAME") == 0)
    errstr << "TDLYFILENAME is not defined " 
        <<  Sas::error("NoTDLYFILENAME");
  else
    sDelaysFilename = std::getenv("TDLYFILENAME");
   
  DataSet * DelaysDataSet = dataSetServer->open( sDelaysFilename, DataSet::Read );
  //DataSet * DelaysDataSet = dataSetServer->open( sDelaysFilename, Dal::Read );

  if (DelaysDataSet->hasAttribute("TDLYVER")) {

    errstr << " TDLYVER keyword has been read from " << sDelaysFilename <<  " file " << Sas::warning("delaysVersionRead");
    //sDelayVer = DelaysDataSet->attribute("TDLYVER")->asString(); 
    sDelayStream << DelaysDataSet->attribute("TDLYVER")->asString(); 
    sDelayVer = sDelayStream.str();

    }
  else
    {

    errstr << "Unable to read TDLYVER keyword in " << sDelaysFilename << " file " << Sas::error("NoTDLYVER");

    }

    dataSetServer->close(DelaysDataSet);
    *dataSet->addAttribute("TDLYVER","Version of g/s delays file")=sDelayVer; 
  // End SPR x2

  // Create the data table and set its attributes
  tab = dataSet->addTable("SCRTX1", size);
  *tab->addAttribute("EXTVER", "Extension Version") = 1;
  *tab->addAttribute("TELESCOP", "XMM Mission") = "XMM";
  *tab->addAttribute("INSTRUME", "Spacecraft Instrument") = "SC  ";
  *tab->addAttribute("DATATYPE", "Spacecraft reconstructed time correlation data") = "RCTIMC.EL";
  *tab->addAttribute("DATE-OBS", "Start time of revolution") = Start.toString(0);
  *tab->addAttribute("DATE-END", "End time of revolution") = End.toString(0);

  // FIXME: TFORM(OBTFINE) == I, should be U for unsigned 16 bit int.
  // FIXME: TFORM(*) missing magnitude, i.e. a column with 1 32-bit signed int should be 1J, not J

  std::cout << "Creating columns1" << std::endl;
  
  // Create the columns
  tab->addColumn("FRAME", Column::Int32, "", "COUNTER");
  tab->addColumn("OBTCOARS", Column::Int32, "Coarse COBT", "SECOND");
  tab->addColumn("OBTFINE", Column::Uint16, "Fine COBT", "        ");
  tab->addColumn("UTCDAY", Column::Int16, "UTC (day) for COBT", "        ");
  // Create the columns
  std::cout << "Creating columns2" << std::endl;
  tab->addColumn("UTCMILLI", Column::Int32, "UTC (ms of day) for COBT", "        ");
  tab->addColumn("UTCMICRO", Column::Int16, "UTC (us of day) for COBT", "        ");
  std::cout << "Creating columns3" << std::endl;
  tab->addColumn("UTCFLAG", Column::Int16, "UTC quality flag", "        ");
  std::cout << "Creating columns3a" << std::endl;
  tab->addColumn("ERTIME", Column::DString, "Corrected ERT", "        ", timesize);
  std::cout << "Creating columns4" << std::endl;
  tab->addColumn("GSID", Column::DString, "Ground Station identifier", "        ", gsidsize);
  tab->addColumn("ERTIMRAW", Column::DString, "Uncorrected ERT", "        ", timesize);
  tab->addColumn("UTCREAL", Column::Real64, "UTC", "        ");
  tab->addColumn("OBTREAL", Column::Real64, "COBT", "        ");
  // Create the columns
  std::cout << "Creating columns" << std::endl;
  tab->addColumn("PASS", Column::Int16, "Pass Number", "        ");
  tab->addColumn("GS", Column::Int16, "Ground Station identifier", "        ");
  tab->addColumn("PREDUTC", Column::Real64, "", "        ");
  tab->addColumn("RESIDUAL", Column::Real64, "", "        ");

  std::cout << "Done" << std::endl;

  // Set up the data pointers
  open();

}


void RTX::close() {

  // Close the dataset  
  dataSetServer->close(dataSet);
  dataSet = 0;
  
  // Go through and release memory

}

void RTX::open(Dal::AccessMode mode) {

  if (dataSet == 0) {
    dataSet = dataSetServer->open( getFilename(".FIT"), mode );
    tab     = dataSet->table("SCRTX1");
  }

  frameCol   = tab->column("FRAME");
  obtcCol    = tab->column("OBTCOARS");
  obtfCol    = tab->column("OBTFINE");
  utcdayCol  = tab->column("UTCDAY");
  utcmilCol  = tab->column("UTCMILLI");
  utcmicCol  = tab->column("UTCMICRO");
  utcflagCol = tab->column("UTCFLAG");
  ertCol     = tab->column("ERTIME");
  gsidCol    = tab->column("GSID");
  ertrawCol  = tab->column("ERTIMRAW");
  utcreaCol  = tab->column("UTCREAL");
  obtrCol    = tab->column("OBTREAL");
  passCol    = tab->column("PASS");
  gsCol      = tab->column("GS");
  predERTCol = tab->column("PREDUTC");
  residCol   = tab->column("RESIDUAL");

  // Get the column data
  frameColD   = frameCol->data();
  obtcColD    = obtcCol->data();
  obtfColD    = obtfCol->data();
  utcdayColD  = utcdayCol->data();
  utcmilColD  = utcmilCol->data();
  utcmicColD  = utcmicCol->data();
  utcflagColD = utcflagCol->data();
  ertColD     = ertCol->data();
  gsidColD    = gsidCol->data();
  ertrawColD  = ertrawCol->data();
  utcreaColD  = utcreaCol->data();
  obtrColD    = obtrCol->data();
  passColD    = passCol->data();
  gsColD      = gsCol->data();
  predERTColD = predERTCol->data();
  residColD   = residCol->data();

  // Get pointers to the data
  frameData   = frameColD->int32Data();
  obtcData    = obtcColD->int32Data();
  obtfData    = obtfColD->uint16Data();
  utcdayData  = utcdayColD->int16Data();
  utcmilData  = utcmilColD->int32Data();
  utcmicData  = utcmicColD->int16Data();
  utcflagData = utcflagColD->int16Data();
  utcreaData  = utcreaColD->real64Data();
  obtrData    = obtrColD->real64Data();
  passData    = passColD->int16Data();
  gsData      = gsColD->int16Data();
  predERTData = predERTColD->real64Data();
  residData   = residColD->real64Data();

}

Pass RTX::fitRevolution(bool ignoreBadPoints) {

  // CR x7
  // SPRx24
  std::cout << "Pass RTX:: Ready for Fit whole revolution" << std::endl;
  std::cout << " Parameter ignoreBadPoints" << ignoreBadPoints << std::endl;
  std::cout << "  Ready for initialising Rev as ficticious Pass" << std::endl;

  Pass fit(0, 0, 0);

  // SPRx24
  std::cout << "Pass RTX:: Ready for adding " << tab->rows()
            << " OBT-UTC pairs (if flagged as good or forced) " << std::endl;

  unsigned int i;

  // Generate fit for each pass
  for (i = 0; i < tab->rows(); i++) {
    
    // Add the data points to the current pass
    if (!utcflagData[i]) {
      fit.addPoint(obtrData[i],utcreaData[i]);
    }
    else {
      if (!ignoreBadPoints) {
	fit.addPoint(obtrData[i],utcreaData[i]);
      }
    }
  }
    
  fit.setEnd(i);
  // SPRx24
  std::cout << "Pass RTX:: OBT-UTC pairs added. End set in " << i << std::endl;
  std::cout << "Pass RTX:: Ready for comparing against Predicted ERT and compute residuals  " << std::endl;
  
  // Update predicted ERT and residuals
  for (i = 0; i < tab->rows(); i++) {
    
    predERTData[i] = fit.getPredictedERT(obtrData[i]);
    residData[i] = 0.0;

    if (!utcflagData[i]) {
      residData[i] = utcreaData[i] - predERTData[i];
    }
    else {
      if (!ignoreBadPoints) {
	residData[i] = utcreaData[i] - predERTData[i];
      }
    }

  }

  // SPRx24
  std::cout << "Pass RTX:: Exiting, now we have residuals (not fit yet). " << i << std::endl;
  std::cout << "   i.e. We have OBT-UTC and residuals from preliminar fit not saved anywhere.  " << i << std::endl;

  return fit;

}

map<int, Pass *> RTX::fitPasses(bool ignoreBadPoints, bool fitEntirePass) {
  
  //
  // Generate per-pass fit 
  //

  map<int, Pass *> passList;
  Pass *currentPass = 0;
  unsigned int i;
  int thisGSID = 0;
  int lastGSID = 0;
  int thisPass = 0;
  int lastPass = 0;
  int stationPoints = 0;
  bool addPoint;
  int fitPoints = 0;

  // CR x7
  // SPRx24
  std::cout << "map Pass RTX:: fitPasses, entering." << std::endl;
  std::cout << "      map Pass RTX:: ignoreBadPoints" << ignoreBadPoints << " fitEntirePass " << fitEntirePass << std::endl;
  
  if (!fitEntirePass) {
    openParameters("stspproc");
    fitPoints = intParameter("passfitpoints");
    // SPRx24
    std::cout << "      map Pass RTX:: Using passfitpoints" << fitPoints << std::endl;
  }

  // Generate fit for each pass
  for (i = 0; i < tab->rows(); i++) {

    thisPass = passData[i];
    thisGSID = gsData[i];
    addPoint = true;

    // If we have switched ground stations
    if (lastGSID != thisGSID) {

      // If this is not the first pass in the dataset
      if (lastGSID == 0) {
        // SPRx24
        std::cout << "      map Pass RTX:: Ready for creating first pass " << std::endl;

	// Create a new pass
	currentPass = new Pass(thisGSID, thisPass, i);
	currentPass->setDataOffset(fitPoints);
	stationPoints = 0;

	// Update the last GS
	lastGSID = thisGSID;

      }
      else {
        // SPRx24
        std::cout << "      map Pass RTX:: Detected g/s h/o. " << std::endl;
        std::cout << "      map Pass RTX:: Closing previous pass. " << std::endl;
        std::cout << "      map Pass RTX:: Previous pass of " 
                  << stationPoints << " points" << std::endl;

	// If the pass is stable enough
	if (stationPoints < fitPoints) {
          // SPRx24
          std::cout << "      map Pass RTX:: pass is too short!!!. " << std::endl;
	  errstr << "Unstable pass, to be fixed later" << Sas::warning("unstable");
          // NEW SPRx15 Enhanced passes verbosity
          std::cout << " Unstable pass, stations points: " << stationPoints
                    << " FitPoints: " << fitPoints 
                    << " - to be fixed later " <<  std::endl;
          // END SPRx15  
	}

	// Set end of current pass
	currentPass->setEnd(i - 1);
	
	// Add pass to the list of passes
	passList[lastPass] = currentPass;
	
	// Create a new pass
        // SPRx24
        std::cout << "      map Pass RTX:: Now creating the newly detected one." << std::endl;
	currentPass = new Pass(thisGSID, thisPass, i);
	currentPass->setDataOffset(fitPoints);
	stationPoints = 0;
	
	// Update the last GS
	lastGSID = thisGSID;
	
      }
	
    }
    else {
      stationPoints++;
      lastGSID = thisGSID;
    }

    // Add the data points to the current pass
    if (addPoint) {
      if (!utcflagData[i]) {
	currentPass->addPoint(obtrData[i],utcreaData[i]);
      }
      else {
	if (!ignoreBadPoints) {
	  currentPass->addPoint(obtrData[i],utcreaData[i]);
	}
      }
    }
  
    // Remember last Pass
    lastPass = thisPass;

  }
  
  // Close the last pass record, if it exists
  if (currentPass) {

    // 
    currentPass->setEnd(i - 1);
    
    // Add pass to the list of passes
    passList[lastPass] = currentPass;
    
  }

  // SPRx24
  std::cout << "map Pass RTX:: All OBT-UTC pairs added to their passes." << std::endl;
  // Update predicted ERT and residuals
  for (i = 0; i < tab->rows(); i++) {
    
    predERTData[i] = passList[passData[i]]->getPredictedERT(obtrData[i]);

    residData[i] = 0.0;

    if (!utcflagData[i]) {
      residData[i] = utcreaData[i] - predERTData[i];
    }
    else {
      if (!ignoreBadPoints) {
	residData[i] = utcreaData[i] - predERTData[i];
      }
    }

  }

  return passList;

}

void RTX::plotRevolution(string title, string subtitle, string extension) {

  // Get data parameters
  vector<string> yparam;
  yparam.push_back(string("RESIDUAL"));
  
  // Plot data
  openParameters("dsplot");
  setParameter("plotter","grace");
  setParameter("title", title);
  setParameter("subtitle", subtitle);
  setParameter("withtable", true);
  setParameter("table", getFilename(".FIT") + ":SCRTX1");
  setParameter("withx", true);
  setParameter("x", string("OBTREAL"));
  setParameter("withy", true);
  setParameter("y", yparam);
  // SPRx27 
  //setParameter("withscatter", true);
  setParameter("tofile", true);
  setParameter("file", getFilename(extension));
  task(dsplot);

}

// CR x8
void RTX::plotInputsRevolution(string title, string subtitle, string extension) {

  // Get data parameters
  vector<string> yparam;
  yparam.push_back(string("UTCREAL"));

  // Plot data
  openParameters("dsplot");
  setParameter("plotter","grace");
  setParameter("title", title);
  setParameter("subtitle", subtitle);
  setParameter("withtable", true);
  setParameter("table", getFilename(".FIT") + ":SCRTX1");
  setParameter("withx", true);
  setParameter("x", string("OBTREAL"));
  setParameter("withy", true);
  setParameter("y", yparam);
  setParameter("withscatter", true);
  setParameter("tofile", true);
  setParameter("file", getFilename(extension));
  task(dsplot);

}



void RTX::truncateFile() {

  unsigned int i;

  //
  // Remove entries from outside of revolution
  //

  std::vector<double> time;
  CalOalUtility::fillVector(tab, "UTCREAL", time);
  
  // Update predicted ERT and residuals
  for (i = 0; i < tab->rows(); i++) {

   STime entry = Utils::MJD20002STime(time[i] / 86400.0);

    if (entry >= end) {

      tab->deleteRows(i, tab->rows() - i);
      break;

    }
  }

  for (i = 0; i < tab->rows(); i++) {

    STime entry = Utils::MJD20002STime(time[i] / 86400.0);

    if (entry >= start) {

      tab->deleteRows(0, i);
      break;

    }
  }
}

// NEW SPRx31
void RTX::moveup() {

  unsigned int i;
  Utils::CDSTime cdsT;
  STime old_ert;

  errstr << "Moving nofit data to newly created nofit columns " 
         << Sas::message(Msg::App, Msg::Verbose);

  // Add two new columns for containing no fitted data
  tab->addColumn("UTCNOFIT", Column::Real64, "UTC nofitted", "        ");
  tab->addColumn("OBTNOFIT", Column::Real64, "OBT nofitted", "        ");

  // Create pointer to new columns
  obtnofitCol = tab->column("OBTNOFIT");
  utcnofitCol = tab->column("UTCNOFIT");

  // Get the columns data
  utcnofitColD  = utcnofitCol->data();
  obtnofitColD  = obtnofitCol->data();

  // Get pointers to the data
  utcnofitData  = utcnofitColD->real64Data();
  obtnofitData  = obtnofitColD->real64Data();

  for (i = 0; i < tab->rows(); i++) {
    // First Movement: OBTCoarse and Fine to obtnofit
    //obtnofitData[i]=obtcData[i]+obtfData[i]/65535.0;
    obtnofitData[i]=obtrData[i];
      
    // Second Movement: UTCday/micro/milli to UTCnofit
    cdsT.Days = utcdayData[i];
    cdsT.Milliseconds = utcmilData[i];
    cdsT.Microseconds = utcmicData[i];
    old_ert = Utils::cds2STime(cdsT);
    utcnofitData[i]=Utils::STime2MJD2000(old_ert) * 86400.0;
    // Or alternatively, string ERTime to UTCnofit
    //CalOalUtility::stringColumnRowSet(utcnofitColD, currentRow,old_ert.toString(6));

  }
 
  // End of creation of new columns

}
// END SPRx31

void RTX::cleanup() {

  unsigned int i;

  for (i = 0; i < tab->rows(); i++) {

    // Update the frame counter
    frameData[i] = i + 1;

    // Get the time
    STime entry = Utils::MJD20002STime(utcreaData[i] / 86400.0);

    // Update the ERT field
    CalOalUtility::stringColumnRowSet(ertColD, i, entry.toString(6));    

    // Update the UTC fields
    Utils::CDSTime cdsT = Utils::STime2cds(entry);
    utcdayData[i] = cdsT.Days;
    utcmilData[i] = cdsT.Milliseconds;
    utcmicData[i] = cdsT.Microseconds;
  
  }

  tab->deleteColumn("UTCREAL");
  tab->deleteColumn("OBTREAL");
  tab->deleteColumn("PASS");
  tab->deleteColumn("GS");
  tab->deleteColumn("PREDUTC");
  tab->deleteColumn("RESIDUAL");

}

string RTX::getFilename(string extension) {
  
  stringstream s; 
  
  s << "RTX_";
  
  s.width(4); s.fill('0');
  s << revolution << "_";
  
  s.width(4); s.fill('0');
  s << version << extension;
  
  return s.str();

}
