// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// CR x3 jcv 8-Oct-08 Additional logging info on extracted data window
// SPRx20 jcv 17-Dec-09 Added the possibility of NOT fitting anything
// SPRx38 jcv 10-Feb-12 Fixed typo in trace with pre/postwin 

#include <iostream>
#include <sstream>
#include <fstream>

#include <vector>

#include "DUExtractor.h"
#include "NCTRSProcessor.h"
#include "PacketProcessor.h"
#include "STSPProcessor.h"

#include "stspproc.h"
#include "Revolution.h"

#include "errstr.h"
#include "ExceptionErrorHandler.h"

///////////////////////
//
//
//
//
void stspproc () {
  
  Revolution rev;

  // NEW SPRx20
  bool besmart = booleanParameter("performfit");
  // END SPRx20

  int rev_number = intParameter("revolution");

  // Negative revolution number means add the (-ve) number to the current
  // revolution number, to offset into the past. Clvr
  if (rev_number <= 0) {

    rev_number = rev.getRev(STime("now")) + rev_number;

  }

  int ver_number = intParameter("version");

  stringstream revString; 
  revString.width(4); revString.fill('0');
  revString << rev_number;

  stringstream verString; 
  verString.width(4); verString.fill('0');
  verString << ver_number;

  STSPProcessor proc = STSPProcessor(rev_number,ver_number);
  proc.Start = rev.getStart(rev_number);
  proc.End   = rev.getEnd(rev_number);

  errstr << "Rev Limits: " << proc.Start << " - " << proc.End << Sas::message(Msg::App, Msg::Sparse);
  
  // Get data for Rev+-18 hours
  // PARAM directory for raw tm
  double prewindow = realParameter("prewindow");
  double postwindow = realParameter("postwindow");

  // CR x3 Additional loggging info on extracted data window
  errstr << "Prewindowing: Extracting from: " << (proc.Start-prewindow) << Sas::message(Msg::App, Msg::Sparse);
  // SPRx38 Fixed small typo
  errstr << "Postwindowing: Extracting up: " << (proc.End+postwindow) << Sas::message(Msg::App, Msg::Sparse);

  stringstream output;
 
  try {

    SetExceptionErrorHandler x; 

    DUExtractor myExtractor( proc.Start - prewindow, proc.End + postwindow );
    
    NCTRSProcessor * myNCTRSProcessor = new NCTRSProcessor(&proc, NCTRSProcessor::VC0);
    PacketProcessor * myPacketProcessor = new PacketProcessor(&proc);
    
    myExtractor.registerProcessor(myNCTRSProcessor);
    myNCTRSProcessor->registerProcessor(myPacketProcessor);


    // NEW CR x12 Refactored DU processing
    //myExtractor.processDU();
    myExtractor.loopDU();

    proc.createRTX(); 

    //proc.correctSTSPs(output);
    // NEW SPRx20
    if (besmart)
     {
        errstr << "running smart: will fit data" << std::endl;
        proc.correctSTSPs(output);
     }

  }
  catch(SasException::Error e){
    
    output << "ERROR: There was an error in Revolution "
	   << revString.str()
	   << ": "
	   << e.code 
	   << ": "
	   << e.message
	   << std::endl;
      
    errstr << "ERROR: There was an error "
	   << e.code 
	   << ": "
	   << e.message
	   << Sas::warning("ProcError");

  }
  catch(SasException::Fatal e){

    output << "ERROR: There was a fatal error in Revolution "
	   << revString.str()
	   << ": "
	   << e.code 
	   << ": "
	   << e.message
	   << std::endl;

    errstr << "ERROR: There was a fatal error "
	   << e.code 
	   << ": "
	   << e.message
	   << Sas::warning("ProcFatal");
  }
  catch(...){

    output << "ERROR: There was a unknown fatal error in Revolution "
	   << revString.str()
	   << ": "
	   << std::endl;

    Sas::warning("UnknownError", "An error was detected, but I can't tell which one. At this stage this is fatal.");

  }    

  stringstream fname; fname << "RTX_"  << revString.str() << "_" << verString.str() << ".log"; 

  ofstream file;
  file.open(fname.str().c_str());
  file << output.str();
 
}
