// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// SPRx19 jcv 4-Dec-09 Completly rebuilt for TCX analysis
// SPRx19b jcv 18-Dec-09 Fixed output problem
// SPRx22 jcv 4-Feb-2010 gcc4 migration

#include <iostream>

#include "tcxdump.h"
#include "TCX.h"
#include "CalOalUtility.h"
#include "Utils.h"
#include "STime.h"

inline double convertToDouble(const std::string& s);

///////////////////////
//
//
//
//
void tcxdump () {

  // Get the filename from parameter  
  string infile = stringParameter("file");
  STime start = stringParameter("starttime");
  STime end = stringParameter("endtime");

  // Chop the filename for getting the rev and the obs
  int revno = atoi(infile.substr(0,4).c_str());
  int obsid = atoi(infile.substr(5,10).c_str());

  // Create the TCX structure, 
  TCX tcsfile(TCXfiles::TCX, revno, obsid);

  // Other initialisations
  unsigned lastObtCoarse = 0, nWraps = 0, nskips = 0;
  bool firstRecord = false;
  STime _utcBaseTime;

  // Read the file
  // Loop on records
  for (unsigned int i = 0; i < tcsfile.tab->rows(); i++ ) {

    // Check within limits
    STime thisRec(CalOalUtility::stringColumnRow(tcsfile.ertColD, i));
    if ( (thisRec < start) || (thisRec > end ) ) { 
       errstr << "Skipping record at "
                << thisRec.toString()
                << Sas::message(Msg::App, Msg::Verbose);
       continue;
    }

    int TCSOBTC = tcsfile.obtcData[i];
    int TCSOBTF = tcsfile.obtfData[i];

    if (!firstRecord) {

      firstRecord = true;

      //        get very first UTC field and establish _utcBaseTime
      _utcBaseTime = STime(1958, 1, 1 + tcsfile.utcdayData[i]);

      errstr << "Processing first record, "
         << thisRec.toString() 
         << " UTC base "
         << _utcBaseTime 
         << Sas::message(Msg::App, Msg::Verbose);

      // Init this for the first record
      lastObtCoarse = TCSOBTC;
    }

    // Some additional checks
    if (tcsfile.utcdayData[i] <= 0 || tcsfile.utcflagData[i]) {
            errstr << "Skipping nonreliable records!!!" 
                 << Sas::message(Msg::App, Msg::Verbose);
            nskips++;
            continue;   //      ignore entries with
                        //              * UTCDAY==0
                        //              * UTCFLAG!=0
                        //      note: some early ODFs e.g. 0078/0124100101 have
                        //      UTCDAY==0 but UTCFLAG==0 so, UTCFLAG is not
                        //      a reliable quality indicator
            }

    // Get OBT value
    //
    int diff = TCSOBTC - lastObtCoarse;
    errstr << "Processing record #" << i 
           << " OBTCOARSE " << TCSOBTC
           << " (last OBTCOARSE " << lastObtCoarse << "," 
           << " diff is  " << diff  << ")" 
           << Sas::message(Msg::App, Msg::Verbose);


    if (diff < 0)
    // SPRx22 gcc4 migration, added not-needed braces
    {
       if (-(double)diff > .9*65536.*65536.)
              {
              //        coarse time must have had a wrap around
              errstr << "Detected apparent wrap around of onboard clock " 
                     << " at record #" << i 
                     << " where OBTCOARSE jumps from "
                     << lastObtCoarse << " to " << TCSOBTC
                     << Sas::message(Msg::App, Msg::Verbose);
              ++nWraps;
              }
            else {
              //        this is an OBT reset 
              errstr << "Detected apparent reset of onboard clock " 
                     << " at record #" << i 
                     << " where OBTCOARSE jumps from "
                     << lastObtCoarse << " to " << TCSOBTC
                     << Sas::message(Msg::App, Msg::Verbose);
              //break;
             }

    // SPRx22 gcc4 migration, added not-needed braces
    }

    // From here, assumed diff is positive

    lastObtCoarse = TCSOBTC;

    //double OBTsimple= (double)TCSOBTC + (double)TCSOBTF / 65536.0; 

    double OBT = (double)TCSOBTC + (double)TCSOBTF / 65536.0 +
           (double)nWraps* 65536. * 65536.; 

    STime ERT(CalOalUtility::stringColumnRow(tcsfile.ertColD, i));

    // Get UTC Value 
    // For TCX files, MET is ERT minus base (see oal/src/TimeCorr.cc)
    STime TEST = ERT - _utcBaseTime;
    double met =  ERT - _utcBaseTime;

    // For TCX files, met is different 
    //STime MET_TCSONLY = STime(1958, 1, 1 + tcsfile.utcdayData[i], 0, 0, 0,
    //  1.e-3 * (double)(tcsfile.utcmilData[i]) + 1.e-6 * (double)(tcsfile.utcmicData[i])) - _utcBaseTime);
    //STime TCSRAWERT(CalOalUtility::stringColumnRow(tcsfile.ertrawColD, i));

    // Get ERT Value 
    double ert = ERT - _utcBaseTime;
 
    // Get the GS id 
    string TCSgsid(CalOalUtility::stringColumnRow(tcsfile.gsidColD, i));

    //std::cout.precision(32);
    std::cout.precision(12);
    std::cout 
	<< "OBT: "
	<< OBT
        // TEST is curious, as it coincides with met and ert 
        // (whem utcbase is substracted)
	<< " met/ert: " 
	<< TEST.modifiedJulianDay().full()
	<< " ERT(MJD): "
	<< ERT.modifiedJulianDay().full()
	<< " ERT: "
	<< ERT.toString()
	<< " GS: "
	<< TCSgsid 
	<< std::endl; 

    errstr << setprecision(12) 
        << "TimeCorrDataPoint: OBT: "
	<< OBT
	<< " met(minus utcbase): " 
	<< met 
	<< " ert(minus utcbase): "
	<< ert
	<< " ERT: "
	<< ERT.toString()
	<< " GS: "
	<< TCSgsid 
        << Sas::message(Msg::App, Msg::Verbose);
    

  // Continue loop on records
  }
  // We have finished loop on records
  errstr << "Nonreliable records: " << nskips
                 << Sas::message(Msg::App, Msg::Verbose);
  errstr << "Number of Wraps arounds: " << nWraps
                 << Sas::message(Msg::App, Msg::Verbose);

}

inline double convertToDouble(const std::string& s)
{
	std::istringstream i(s);

	double x;
	
	i >> x;	

	return x;

} 

